/*
 * Copyright (c) 2019-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "Spheres.h"

namespace visrtx {

Spheres::~Spheres()
{
  cleanup();
}

void Spheres::commit()
{
  Geometry::commit();

  cleanup();

  m_radius = getParamObject<Array1D>("vertex.radius");

  m_vertex = getParamObject<Array1D>("vertex.position");
  m_vertexColor = getParamObject<Array1D>("vertex.color");
  m_vertexAttribute0 = getParamObject<Array1D>("vertex.attribute0");
  m_vertexAttribute1 = getParamObject<Array1D>("vertex.attribute1");
  m_vertexAttribute2 = getParamObject<Array1D>("vertex.attribute2");
  m_vertexAttribute3 = getParamObject<Array1D>("vertex.attribute3");

  if (!m_vertex) {
    reportMessage(ANARI_SEVERITY_WARNING,
        "missing required parameter 'vertex.position' on sphere geometry");
    return;
  }

  m_vertex->addCommitObserver(this);

  if (hasParam("radius"))
    m_globalRadius = getParam<float>("radius", 0.01f);
  else
    m_globalRadius.reset();

  float globalRadius = m_globalRadius.value_or(0.01f);

  // Calculate bounds //

  m_aabbs.resize(m_vertex->size());

  {
    auto *begin = m_vertex->hostDataAs<vec3>();
    auto *end = begin + m_vertex->size();

    float *radius = nullptr;
    if (m_radius)
      radius = m_radius->hostDataAs<float>();

    size_t sphereID = 0;
    std::transform(begin, end, m_aabbs.begin(), [&](const vec3 &v) {
      const float r = radius ? radius[sphereID++] : globalRadius;
      return box3(v - r, v + r);
    });
  }

  m_aabbs.upload();
  m_aabbsBufferPtr = (CUdeviceptr)m_aabbs.dataDevice();
}

void Spheres::populateBuildInput(OptixBuildInput &buildInput) const
{
  buildInput.type = OPTIX_BUILD_INPUT_TYPE_CUSTOM_PRIMITIVES;

  buildInput.customPrimitiveArray.aabbBuffers = &m_aabbsBufferPtr;
  buildInput.customPrimitiveArray.numPrimitives = m_aabbs.size();

  static uint32_t buildInputFlags[1] = {OPTIX_GEOMETRY_FLAG_NONE};

  buildInput.customPrimitiveArray.flags = buildInputFlags;
  buildInput.customPrimitiveArray.numSbtRecords = 1;
}

GeometryGPUData Spheres::gpuData() const
{
  auto retval = Geometry::gpuData();
  retval.type = GeometryType::SPHERE;

  auto &sphere = retval.sphere;

  sphere.centers = m_vertex->deviceDataAs<vec3>();
  sphere.radii = nullptr;
  if (m_radius)
    sphere.radii = m_radius->deviceDataAs<float>();
  sphere.radius = m_globalRadius.value_or(0.01f);

  populateAttributePtr(m_vertexAttribute0, sphere.vertexAttr[0]);
  populateAttributePtr(m_vertexAttribute1, sphere.vertexAttr[1]);
  populateAttributePtr(m_vertexAttribute2, sphere.vertexAttr[2]);
  populateAttributePtr(m_vertexAttribute3, sphere.vertexAttr[3]);
  populateAttributePtr(m_vertexColor, sphere.vertexAttr[4]);

  return retval;
}

int Spheres::optixGeometryType() const
{
  return OPTIX_BUILD_INPUT_TYPE_CUSTOM_PRIMITIVES;
}

void Spheres::cleanup()
{
  if (m_vertex)
    m_vertex->removeCommitObserver(this);
}

} // namespace visrtx
